// Copyright (c) 2021 Dominic Msters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
extern "C" {
  #include <glad/glad.h>
  #include <GLFW/glfw3.h>
  #include "libs.h"
}

#define WINDOW_WIDTH_DEFAULT 1280
#define WINDOW_HEIGHT_DEFAULT WINDOW_WIDTH_DEFAULT/16*9

#define GLFW_PLATFORM_INPUT_MOUSE_X (inputsource_t)0xFF
#define GLFW_PLATFORM_INPUT_MOUSE_Y (inputsource_t)0xFE

class GLFWEngine {
  private:
    char **args;
    int32_t argc;

  public:
    GLFWwindow *window;

    /**
     * Construct a new GLFW Engine instance.
     * 
     * @param argc Count of arguments.
     * @param args Array of strings for the arguments provided to the engine.
     */
    GLFWEngine(int32_t argc, char **args);

    /**
     * Initializes the GLFW engine.
     * 
     * @return 0 if success, everything else is a failure.
     */
    virtual int32_t init(void);

    /**
     * Tick the game engine by delta amount.
     * 
     * @param delta 
     * @return 0 if success, everything else is a failure.
     */
    int32_t update(float delta);

    virtual ~GLFWEngine(void) override;
};


/**
 * Entry of the program
 * 
 * @param argc Count of args in the args array.
 * @param args Args provided to us by the parent O.S.
 * @return 0 if success, anything else for failure.
 */
int32_t main(int32_t argc, char **args);

/**
 * Resize callbacks.
 * 
 * @param window Window that was resized.
 * @param width New window width.
 * @param height New window height.
 */
void glfwOnResize(GLFWwindow *window, int32_t width, int32_t height);

/**
 * Keyboard Input callbacks.
 * 
 * @param window Window that was resized.
 */
void glfwOnKey(GLFWwindow *window,
  int32_t key, int32_t scancode, int32_t action, int32_t mods
);

/**
 * Callback for generic GLFW errors.
 * 
 * @param error Error code/descriptor.
 * @param description String representation of the error.
 */
void glfwOnError(int error, const char* description);

/**
 * Event callback for mouse cursor movement.
 * 
 * @param window 
 * @param x 
 * @param y 
 */
void glfwOnCursor(GLFWwindow *window, double x, double y);


/**
 * Get the game engine specific input source for a given GLFW Key code.
 * 
 * @param key Key to get the input source for.
 * @return The input source. 
 */
inputsource_t glfwGetInputSourceForKey(int32_t key);