// Copyright (c) 2021 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
extern "C" {
  #include "../util/math.h"
}
#include "../libs.hpp"


/**
 * Input source identifier. It's up to the platform itself to decide what the
 * hell this number refers to. For most platforms it will be an input, such as a
 * keyboard scancode or a (pad number * button count) + button.
 */
typedef uint8_t inputsource_t;

/**
 * Input Bind, a specific action bind reference for the game engine to use.
 * e.g. "Jump" or "Walk Forward".
 */
typedef uint8_t inputbind_t;

/**
 * Value that represents the state of an input. Defined as 0-1 where 0 is set
 * to be completely off / netural state, and 1 is completely on / full state.
 */
typedef float inputval_t;

/** Debug Inputs */
#define INPUT_NULL      (inputbind_t)0x00

/** Additional sources */
#define INPUT_MOUSE_X   (inputbind_t)0x10
#define INPUT_MOUSE_Y   (inputsource_t)0x11

namespace Dawn {
  class Input {
    private:
      std::map<inputbind_t, std::vector<inputsource_t>> binds;
      std::map<inputbind_t, inputval_t> valsLeft;
      std::map<inputbind_t, inputval_t> valsRight;
      std::map<inputsource_t, inputval_t> buffer;
      bool left;
  
    public:
      /**
       * Initializes the input manager.
       */
      Input(void);

      /**
       * Tick the input manager.
       */
      void update(void);

      /**
       * Binds the given input binding to the input source. Essentially allowing any
       * time we fetch the state of bind, we will read the value from source.
       * 
       * @param source The source that is being bound.
       * @param bind The binding to bind against.
       */
      void bind(inputsource_t source, inputbind_t bind);

      /**
       * Unbind a previously bound input source from a binding.
       * 
       * @param source The source that is being unbound.
       * @param bind The binding to unbind from.
       */
      void unbind(inputsource_t source, inputbind_t bind);

      /**
       * Set the state of an input.
       * 
       * @param source Source to set.
       * @param value Value to set.
       */
      void setValue(inputsource_t source, float value);

      /**
       * Returns the raw input value as a float between 0 and 1. For digital
       * (buttons) this will typicall be 0 or 1 only. Other analogue inputs will
       * have anywhere within the range.
       * 
       * @param binding The previously bound input binding.
       * @return Input state of the axis.
       */
      inputval_t getAxis(inputbind_t bind);

      /**
       * Is the current input "down", being pressed, being moved, not in a state
       * of rest.
       * 
       * @param binding The previously bound input binding.
       * @return True if the input vector is non-zero.
       */
      bool isDown(inputbind_t binding);

      /**
       * Is the current input "up", in a state of rest, not being actioned,
       * moved.
       * 
       * @param binding The previously bound input binding.
       * @return True if input vector is zero
       */
      bool isUp(inputbind_t binding);

      /**
       * Returns true on the first tick that an input was actioned/downed.
       * 
       * @param binding The previously bound input binding.
       * @return True if the input vector was non-zeroed this tick but not last. 
       */
      bool isPressed(inputbind_t binding);

      /**
       * Returns true on the first tick that an input was released/upped.
       * 
       * @param binding The previously bound input binding.
       * @return True if the input vector was zeroed this tick but not last. 
       */
      bool isReleased(inputbind_t binding);

      /**
       * Destroy the input manager and cleanup.
       */
      ~Input(void);
  };
}